import matplotlib.pyplot as plt
import numpy as np
import logging
from .utility import get_xlims

logger = logging.getLogger('laff')

def plot_all_break_fits(data, model_fits, broken_powerlaw):
    """Plot all powerlaw break fits."""

    logger.debug("Running plot_all_break_fits()")

    minval, maxval = np.log10(data['time'].iloc[0]), np.log10(data['time'].iloc[-1])
    constant_range = np.logspace(minval, maxval, num=250)

    fig, axs = plt.subplots(6, sharex=True, sharey=True)
    fig.subplots_adjust(hspace=0, left=0, bottom=0, right=1, top=1)
    
    for fit, ax in zip(model_fits, axs):
        fit = list(fit[0])
        ax.errorbar(data.time, data.flux,
                xerr=[-data.time_nerr, data.time_perr], \
                yerr=[-data.flux_nerr, data.flux_perr], \
                marker='', linestyle='None', capsize=0, zorder=1, color='k')

        fittedContinuum = broken_powerlaw(constant_range, fit)
        ax.plot(constant_range, fittedContinuum, color='c')

        n = int((len(fit)-2)/2)
        
        if n > 0:
            for xpos in fit[n+1:-1]:
                ax.axvline(x=xpos, color='grey', linestyle='--', linewidth=0.5, zorder=0)
        ax.loglog()
        
    plt.xlim(get_xlims(data))
    plt.margins(x=0, y=0)
    plt.show()




    return


from .laff import fitGRB, fitAfterglow, plotGRB, set_logging_level, findFlares, fitFlares
from .lightcurve_import import lcimport
from .utility import STAT_NAMES_CONTINUUM, PAR_NAMES_FLARE, PAR_NAMES_CONTINUUM
from .fit_bat import fitPrompt, plotPrompt


from .fit_afterglow import (
 broken_powerlaw,
 find_afterglow_fit,
 calculate_afterglow_fluence
)

from .fit_flare import (
    fred_flare,
    gaussian_flare,
    flare_fitter,
    improved_end_time
)


import numpy as np
import logging
from scipy.odr import ODR, Model, RealData

from ..utility import calculate_fit_statistics, calculate_fluence

logger = logging.getLogger('laff')

#################################################################################
### AFTERGLOW MODEL
#################################################################################

def broken_powerlaw(params, x):
    x = np.array(x)

    if type(params) in (list, np.ndarray):
        n = int((len(params)-2)/2)
        slopes = params[0:n+1]
        breaks = params[n+1:-1]
        norm   = params[-1]
    elif type(params) is dict:
        n      = params['break_num']
        slopes = params['slopes']
        breaks = params['breaks']
        norm   = params['normal']
    else:
        logger.critical('Input params not accepted type.')
        raise TypeError(f'params is not dict/list -> {type(params)}')

    mask = []

    for i in range(n):
        mask.append(x > breaks[i])

    if n >= 0:
        model = norm * (x**(-slopes[0]))
    if n >= 1:
        model[np.where(mask[0])] = norm * (x[np.where(mask[0])]**(-slopes[1])) * (breaks[0]**(-slopes[0]+slopes[1]))
    if n >= 2:
        model[np.where(mask[1])] = norm * (x[np.where(mask[1])]**(-slopes[2])) * (breaks[0]**(-slopes[0]+slopes[1])) * (breaks[1]**(-slopes[1]+slopes[2]))
    if n >= 3:
        model[np.where(mask[2])] = norm * (x[np.where(mask[2])]**(-slopes[3])) * (breaks[0]**(-slopes[0]+slopes[1])) * (breaks[1]**(-slopes[1]+slopes[2])) * (breaks[2]**(-slopes[2]+slopes[3]))
    if n >= 4:
        model[np.where(mask[3])] = norm * (x[np.where(mask[3])]**(-slopes[4])) * (breaks[0]**(-slopes[0]+slopes[1])) * (breaks[1]**(-slopes[1]+slopes[2])) * (breaks[2]**(-slopes[2]+slopes[3])) * (breaks[3]**(-slopes[3]+slopes[4]))
    if n >= 5:
        model[np.where(mask[4])] = norm * (x[np.where(mask[4])]**(-slopes[5])) * (breaks[0]**(-slopes[0]+slopes[1])) * (breaks[1]**(-slopes[1]+slopes[2])) * (breaks[2]**(-slopes[2]+slopes[3])) * (breaks[3]**(-slopes[3]+slopes[4])) * (breaks[4]**(-slopes[4]+slopes[5]))

    return model

def broken_powerlaw_wrapper(params, x):
    """Wrapper for ODR to fit log10 break times. Intended for internal use only."""

    n = int((len(params)-2)/2)
    params = list(params)
    params[n+1:-1] = [10**val for val in params[n+1:-1]]

    return broken_powerlaw(params, x)
    

#################################################################################
### SCIPY ODR FITTING
#################################################################################

def odr_fitter(data, input_par, convert_to_std=1.0, t_max=0):
    """Fitting function for powerlaw."""
    data = RealData(data.time, data.flux, sx=data.time_perr*convert_to_std, sy=data.flux_perr*convert_to_std)
    model = Model(broken_powerlaw_wrapper)
    n = int((len(input_par)-2)/2)

    odr = ODR(data, model, beta0=input_par)
    output = odr.run()

    if output.info != 1:
        # and all(output.beta[n+1:-1] > 0):
        # and all(output.beta[n+1:-1] < t_max):
        i = 1
        while output.info != 1 and i < 100:
            print(i)
            # and all(output.beta[n+1:-1] > 0):
            # and all(output.beta[n+1:-1] < t_max):
            # print(output.beta[0:n+1])
            # print(output.beta[0:n+1] > -0.6)
            # print(all(output.beta[0:n+1] > -0.6))
            # print(t_max)
            # print('restart')
            # require 0 < break_time < 
            # require 
            output = odr.restart()
            i += 1
            
    return output.beta, output.sd_beta


def find_afterglow_fit(data, conversion_to_std):

    data_start, data_end = data['time'].iloc[0], data['time'].iloc[-1]

    model_fits = []

    logger.debug('breaknum : fit_par / fit_err / fit_stats')
    for breaknum in range(0, 6):

        slope_guesses = [1.0] * (breaknum+1)
        break_guesses = list(np.linspace(np.log10(data_start), np.log10(data_end), num=breaknum+2))[1:-1]
        normal_guess  = [data['flux'].iloc[0] * data['time'].iloc[0]]
        input_par = slope_guesses + break_guesses + normal_guess

        fit_par, fit_err = odr_fitter(data, input_par, convert_to_std=conversion_to_std, t_max=data_end)
        
        if list(fit_par[breaknum+1:-1]) != sorted(list(fit_par[breaknum+1:-1])):
            fit_par[breaknum+1:-1] = sorted(list(fit_par[breaknum+1:-1]))

        fit_par[breaknum+1:-1] = [10**x for x in fit_par[breaknum+1:-1]]
        
        fit_stats = calculate_fit_statistics(data, broken_powerlaw, fit_par)

        model_fits.append([fit_par, fit_err, fit_stats])

    best_fit, best_err, best_stats = min(model_fits, key=lambda x: x[2]['deltaAIC'])
    breaknum = int((len(best_fit)-2)/2)

    logger.debug('brknum\t%s', breaknum)
    logger.debug('slopes\t%s', list(round(x,2) for x in best_fit[0:breaknum+1]))
    logger.debug('breaks\t%s', list(round(x,2) for x in best_fit[breaknum+1:-1]))
    logger.debug('norm\t%s', best_fit[-1])
    # logger.debug('%s', best_err)
    # logger.debug('%s', best_stats)
    logger.info('Afterglow fitted with %s breaks.', breaknum)

    return best_fit, best_err, best_stats, breaknum


#################################################################################
# CALCULATE FLUENCE
#################################################################################

def calculate_afterglow_fluence(data, breaknum, break_times, fit_par, count_to_flux_conversion):

    integral_boundaries = [data.iloc[0].time, *break_times, data.iloc[-1].time]

    # afterglow_fluence = np.sum([calculate_fluence(broken_powerlaw, fit_par, integral_boundaries[i], integral_boundaries[i+1]) for i in range(breaknum)])
    # afterglow_fluence *= count_to_flux_conversion
    # logger.info('Afterglow fluence calculated as %s', afterglow_fluence)

    afterglow_fluence = 0
    return afterglow_fluence


import matplotlib.pyplot as plt
import logging
import numpy as np
from .utility import get_xlims
from ..modelling import broken_powerlaw, fred_flare

logger = logging.getLogger('laff')



def plotGRB(data, grb_fit, save_path=None, show_fig=True, **plot_components):
    """Plot the results of LAFF fit.

    Take the output of a LAFF run and plot the flares, continuum and lightcurve
    of a GRB.

    Args:
        data (pd.Dataframe): GRB lightcurve dataframe.
        grb_fit (dict): the output of laff.fitGRB containing all the fit data.
        save_path (string): default do not save, if provided it will save to given path.
        show (bool): defaults to true, if false do not call plt.show().
        **kwargs:
            lightcurve, continuums, flares, total_model, powerlaw_breaks

    Returns:
        N/A
        TODO - can I return the plotted object?
    """

    # Light curve.
    if not plot_components.get('lightcurve', False):
        logger.debug("Plottting lightcurve")
        plt.errorbar(data.time, data.flux, xerr=[-data.time_nerr, data.time_perr], 
                     yerr=[-data.flux_nerr, data.flux_perr], marker='', linestyle='None', capsize=0, zorder=1)

        upper_flux, lower_flux = data['flux'].max() * 10, data['flux'].min() * 0.1
        
    # For smooth plotting of fitted functions.
    max, min = np.log10(data['time'].iloc[0]), np.log10(data['time'].iloc[-1])
    constant_range = np.logspace(min, max, num=5000)

    # Continuum model.
    if not plot_components.get('continuum', False):
        logger.debug('Plotting continuum model')
        fitted_continuum = broken_powerlaw(constant_range, grb_fit['continuum']['parameters'])
        plt.plot(constant_range, fitted_continuum, color='c')

    try:
        total_model = fitted_continuum
    except NameError:
        total_model = [0] * len(data.time)

    # Flare models.
    if grb_fit['flares'] is not False:
        for flare in grb_fit['flares']:
            flare_model = fred_flare(constant_range, flare['par'])
            total_model += flare_model
            if not plot_components.get('flares', False):
                plt.plot(constant_range, flare_model, color='tag:green', linewidth=0.6, zorder=3)

    # Total model.
    if not plot_components.get('total_model', False):
        plt.plot(constant_range, total_model, color='tab:orange', zorder=5)

    # Powerlaw breaks.
    if not plot_components.get('powerlaw_breaks', False):
        for x_pos in grb_fit['continuum']['parameters']['breaks']:
            plt.axvline(x=x_pos, color='grey', linestyle='--', linewidth=0.5, zorder=0)

    # Adjust limits.
    plt.ylim(lower_flux, upper_flux)
    plt.xlim(get_xlims(data))
    plt.loglog()
    plt.xlabel("Time (s)")
    plt.ylabel("Flux (units)")

    if save_path is not None:
        plt.savefig(save_path)

    if show_fig:
        plt.show()

    return


def saveGRB(data, grb_fit, save_path):
    """Wrapper function to plot and save the LAFF results without showing them."""
    plotGRB(data, grb_fit, save_path, show=False)
    return


from .utility import calculate_fit_statistics, check_data_input, calculate_fluence
from .utility import PAR_NAMES_CONTINUUM, PAR_NAMES_FLARE, STAT_NAMES_CONTINUUM
from .utility import get_xlims
from .rich_output import plot_all_break_fits


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.signal import savgol_filter, find_peaks
from scipy.ndimage import label
from scipy.optimize import least_squares
from scipy.stats import f

from .modelling import fred_flare

################################################################################
# FUNCTIONS
################################################################################

def fred_flares(params, x):
    # J. P. Norris et al., ‘Attributes of Pulses in Long Bright Gamma-Ray Bursts’, The Astrophysical Journal, vol. 459, p. 393, Mar. 1996, doi: 10.1086/176902.

    x = np.array(x)

    flr_params = [params[i:i + 5] for i in range(0, len(params), 5)]

    total_model = [0.0] * len(x)

    for flr in flr_params:

        t_max, rise, decay, sharpness, amplitude = flr

        flr_model = amplitude * np.exp( -(abs(x - t_max) / rise) ** sharpness)
        flr_model[np.where(x > t_max)] = amplitude * np.exp( -(abs(x[np.where(x > t_max)] - t_max) / decay) ** sharpness)

        total_model += flr_model

    return total_model

def fred_resids(params, x, y):
    return fred_flares(params, x) - y

def simple_line(params, x):
    x = np.array(x)
    amplitude, slope = params
    return amplitude + slope * x

def line_resids(params, x, y):
    return simple_line(params, x) - y

################################################################################
# DATA FILTERING
################################################################################

def fitPrompt(data):

    data, flares = filter_data(data)

    continuum = find_continuum(data, flares)

    flares = fit_flares(data, flares, continuum)

    return {'data': data, 'continuum': continuum, 'flares': flares}


def filter_data(data):

    data['savgol'] = savgol_filter(data['flux'], window_length=31, polyorder=3)

    # Filter out negatives.
    filtered_data = data['savgol'].copy()
    filtered_data = filtered_data[filtered_data > 0]
    avg_positive = np.average(filtered_data)
    
    # Filter out big peaks.
    filtered_data = filtered_data[filtered_data < 3 * avg_positive]
    avg_filter = np.average(filtered_data)

    # Calculate residuals.
    data['savgol_residuals'] = data['savgol'] - 3 * avg_filter
    data['savgol_residuals'] = data['savgol_residuals'].apply(lambda x: max(x, 0))
    
    # Calculate rolling std.
    # Cutoff higher peaks.
    modified_flux = data['flux'].copy()
    prev_valid = None
    for i in range(len(modified_flux)):
        if modified_flux[i] > 3 * avg_filter:
            if prev_valid is not None:
                modified_flux[i] = prev_valid
        else:

            prev_valid = modified_flux[i]
    data['moving_std'] = modified_flux.rolling(window=501, min_periods=1).std()
    data.loc[0, 'moving_std'] = data['moving_std'].iloc[1]

    # Gather all flare region indices from residuals.
    labelled_array, num_features = label(data['savgol_residuals'] > 0)
    intial_flare_regions = []

    for i in range(1, num_features+1):
        indices = np.where(labelled_array == i)[0]
        if not indices[0] == indices[-1]:
            intial_flare_regions.append((indices[0], indices[-1]))

    flare_indices = []

    for a, b in intial_flare_regions:
        peak_index = data['savgol'].iloc[a:b].idxmax()
        amplitude = data['savgol'].iloc[peak_index]

        if amplitude < 1.5 * np.average(data['moving_std'].iloc[a:b]):
            continue
        if len(range(a, b)) <= 1:
            continue

        flare_indices.append((a, b))

    flare_indices = sorted(set(flare_indices))

    return data, flare_indices


def find_continuum(data, flare_indices):

    data_continuum = data.copy()

    # Filter flare regions.
    for a, b in flare_indices:
        data_continuum = data_continuum.drop([x for x in range(a, b+1)])

    input_par = [np.average(data_continuum['flux']), 0.0]
    continuum_fit = least_squares(line_resids, input_par, args=(data_continuum['time'], data_continuum['flux'])).x

    return continuum_fit


################################################################################
# FLARE FITTING
################################################################################

def fit_flares(data, flare_indices, continuum_fit):

    flare_data = data.copy()
    flare_data['flux'] -= flare_data['moving_std']
    flare_data['flux'] = flare_data['flux'].apply(lambda x: max(x, 0))

    flares = []

    prev_aic = np.inf

    for a, b in flare_indices:

        add_flare = True
        count_flare = 1

        found_peaks, properties = find_peaks(flare_data['flux'].iloc[a:b+1], prominence=flare_data['moving_std'].iloc[a])            
        ranked_indices = np.argsort(properties['prominences'])[::-1]

        while add_flare == True:

            peaks = [found_peaks[i]+a for i in ranked_indices[:count_flare]]

            if len(peaks) == 0 and count_flare == 1:
                peaks = [data['savgol'].iloc[a:b].idxmax()]

            if len(peaks) != count_flare:
                add_flare = False
                continue

            input_par = []

            for peak in peaks:

                t_peak    = data['time'].iloc[peak]
                rise      = (data['time'].iloc[peak]-data['time'].iloc[a]) / 4 or (data['time'].iloc[b] - np.average((data['time'].iloc[a], data['time'].iloc[b]))) / 4
                decay     = (data['time'].iloc[b]-data['time'].iloc[peak]) / 2 or (data['time'].iloc[b] - np.average((data['time'].iloc[a], data['time'].iloc[b]))) / 2
                sharp     = min(max(decay/rise, 1), 10)
                amplitude = data['savgol'].iloc[peak]

                input_par.extend((t_peak, rise, decay, sharp, amplitude))
            
            bound_par = count_flare * [data['time'].iloc[a], 0.0,            0.0,            1.0,  0.0   ], \
                        count_flare * [data['time'].iloc[b], 2*(rise+decay), 2*(rise+decay), 10.0, np.inf]

            flare_fit = least_squares(fred_resids, input_par, bounds=bound_par, args=(flare_data['time'], flare_data['flux'])).x

            filt_data = data[data['flux'] != 0]

            chi_2 = np.sum( ((filt_data['flux'] - fred_flares(flare_fit, filt_data['time']))**2) / (filt_data['flux_perr'] ** 2) )
            dof_2 = len(data['flux']) - len(flare_fit)

            if count_flare == 1:
                chi_1 = chi_2
                dof_1 = dof_2
                count_flare += 1
                prev_fit = flare_fit
                continue

            F = ((chi_1 - chi_2) / (dof_1 - dof_2)) / (chi_2/dof_2)
            p_value = 1 - f.cdf(F, dof_1-dof_2, dof_2)
            # print(count_flare, p_value)

            if p_value < 0.0027:
                chi_1 = chi_2
                dof_1 = dof_2
                count_flare += 1
                prev_fit = flare_fit
            else:
                flare_fit = prev_fit
                add_flare = False

        flare_data['flux'] -= fred_flares(flare_fit, data['time'])
        flare_data['flux'] = flare_data['flux'].apply(lambda x: max(x, 0))

        for i in range(0, len(flare_fit), 5):
            flares.append({'indices': (a, b), 'parameters': flare_fit[i:i+5]})

    return flares

    
################################################################################
# PLOTTING
################################################################################

def plotPrompt(prompt_fit, **kwargs):
    """_summary_

    Args:
        data (pd.DataFrame): _description_
        prompt_fit (_type_): _description_
    
    Kwargs:
        grb_name (str):   grb name string to title the plot.
        zero_lines (bool): plot the y=0 dashed lines.
        main_data (bool): plot the main data, default true
        residuals (bool): plot the residuals, default true
        flare_spans (bool): plot the flare spans, default true.
        flare_fit
        total fit
        continuum
        save
        TODO savepath
    """

    data = prompt_fit['data']

    residuals = kwargs.get('residuals', True)

    if residuals:
        fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True, figsize=(10,6), gridspec_kw={'hspace': 0})
    else:
        fig, ax1 = plt.subplots(figsize=(10, 6))

    # Plot title.
    if (grb_name := kwargs.get('grb_name')):
        fig.suptitle(str(grb_name))
    
    # Zero lines.
    if kwargs.get('zero_lines', True):
        ax1.axhline(y=0, linestyle='--', color='grey', linewidth=0.5)
        if residuals:
            ax2.axhline(y=0, linestyle='--', color='grey', linewidth=0.5)

    # Main data points.
    if kwargs.get('main_data', True):
        ax1.errorbar(data['time'], data['flux'], yerr=data['flux_perr'], linestyle='None', marker='.', color='grey', linewidth=0.5, alpha=0.3, zorder=-1)

    # Savgol filter line.
    if kwargs.get('savgol', True):
        ax1.plot(data['time'], data['savgol'], color='tab:green', linewidth=0.5)

    # Residuals.
    if residuals:
        ax2.plot(data['time'], data['savgol_residuals'], linewidth=0.5, color='tab:green')


    total_model = [0.0] * len(data['flux'])

    for flare in prompt_fit['flares']:

        srt, end = flare['indices']

        if kwargs.get('flare_spans', True):
            ax1.axvspan(data['time'].iloc[srt], data['time'].iloc[end], color='b', alpha=0.2)
            if residuals:
                ax2.axvspan(data['time'].iloc[srt], data['time'].iloc[end], color='b', alpha=0.2)

        flare_model = fred_flares(flare['parameters'], data['time'])
        total_model += flare_model

        if kwargs.get('flare_fit', True):
            ax1.plot(data['time'], flare_model, color='m', linewidth=0.5, linestyle='--')
            if residuals:
                ax2.plot(data['time'], flare_model, color='m', linewidth=0.5, linestyle='--')

    continuum = simple_line(prompt_fit['continuum'], data['time'])
    total_model += continuum

    if kwargs.get('continuum'):
        ax1.plot(data['time'], continuum, color='m', linewidth=0.5)

    if kwargs.get('total_fit', True):
        ax1.plot(data['time'], total_model, color='black', linewidth=0.7)


    plt.xlabel('Time since trigger (seconds)')
    plt.ylabel('Count rate (/seconds)')

    if (save_path := kwargs.get('save')):
        plt.savefig(save_path + grb_name + '.png', bbox_inches='tight')

    if kwargs.get('show', True):
        plt.show()


import numpy as np
import pandas as pd
import scipy.integrate as integrate
import math
import logging

logger = logging.getLogger('laff')

PAR_NAMES_FLARE = ['t_start', 'rise', 'decay', 'amplitude']
PAR_NAMES_CONTINUUM = ['break_num', 'slopes', 'slopes_err', 'breaks', 'breaks_err', 'normal', 'normal_err']
STAT_NAMES_CONTINUUM = ['chisq', 'rchisq', 'n', 'npar', 'dof', 'deltaAIC']

def calculate_fit_statistics(data, model, params, temp_flare_shell=False):
    
    if temp_flare_shell:
        fitted_model = model(np.array(data.time), params)
    else:
        fitted_model = model(params, np.array(data.time))

    chisq = np.sum(((data.flux - fitted_model) ** 2) / (data.flux_perr ** 2))
    
    n = len(data.time)
    npar = len(params)
    dof = n - npar
    r_chisq = chisq / dof

    deltaAIC = (2 * npar) + (n * np.log(r_chisq))
    deltaAIC = deltaAIC if deltaAIC != -np.inf else np.inf # negative infinity check
        
    return {'chisq': chisq, 'rchisq': r_chisq, 'n': n, 'npar': npar, 'dof': dof, 'deltaAIC': deltaAIC}

def check_data_input(data):

    # if not isinstance(data, pd.DataFrame):
        # raise TypeError(f"Invalid input data type. Should be pandas dataframe.")

    if len(data) < 3:
        logger.critical("Too short.")
        raise ValueError(f"Data too short, require at least 3 data points - given data is {len(data)}")

    if not data.shape[1] == 4 and not data.shape[1] == 6:
        raise ValueError("Expected dataframe of shape (X, 4) or (X, 6)")

    # Check column names.
    expected_columns = ['time', 'time_perr', 'time_nerr', 'flux', 'flux_perr', 'flux_nerr']
    if data.shape[1] == 4:
        data.columns = ['time', 'time_perr', 'flux', 'flux_perr']
        data['time_nerr'] = data['time_perr']
        data['flux_nerr'] = data['flux_perr']
        data.columns = expected_columns
    elif data.shape[1] == 6:
        data.columns = expected_columns
    else:
        raise ValueError(f"Expected dataframe with 4 or 6 columns - got {data.shape[1]}.")
    
    data = data.reset_index(drop=True)


    logger.debug('Data input is good.')

    return True

def calculate_fluence(model, params, start, stop, count_ratio):
    """Given some model and range, calculate the fluence."""

    range = np.logspace(np.log10(start), np.log10(stop), num=2500)
    fitted_model = model(range, params)
    fluence = integrate.trapezoid(fitted_model, x=range)
    logger.debug('Fluence from %s to %s is %s', start, stop, fluence)

    return fluence * count_ratio

def get_xlims(data):
    """Calculating xlimits for a log graph.

    Calculate the xlimits for a log graph so there's no large spaces on either
    end. I'm probably missing an easy bit of math. If the value is too close,
    i.e. log10(time) = 2.05, then substracting 10**2 is too much, so included an
    edge case.

    Args:
        data (pd.Dataframe): GRB lightcurve dataframe

    Returns:
        float: lower x-axis limit
        float: upper x-axis limit
        _type_: _description_
    """
    lowest_time_val = data['time'].iloc[0] + data['time_nerr'].iloc[0]
    lowest_time_val_power = math.floor(np.log10(lowest_time_val))
    lowest_time_val_power = lowest_time_val_power-1 if lowest_time_val_power - math.floor(lowest_time_val) < 0.2 else lowest_time_val_power
    lower_lim = lowest_time_val - 10**lowest_time_val_power

    highest_time_val = (data['time'].max() + data['time_perr'].iloc[len(data.time)-1])
    upper_lim = highest_time_val + 10**(math.floor(np.log10(highest_time_val)))

    return lower_lim, upper_lim



import numpy as np
import logging
import emcee
from intersect import intersection
from ..utility import calculate_fit_statistics
from ..modelling import broken_powerlaw

logger = logging.getLogger('laff')

#################################################################################
### FRED MODEL
#################################################################################

def fred_flare(x, params):
    # J. P. Norris et al., ‘Attributes of Pulses in Long Bright Gamma-Ray Bursts’, The Astrophysical Journal, vol. 459, p. 393, Mar. 1996, doi: 10.1086/176902.
    
    x = np.array(x)
    t_max = params[0]
    rise = params[1]
    decay = params[2]
    sharpness = params[3]
    amplitude = params[4]

    model = amplitude * np.exp( -(abs(x - t_max) / rise) ** sharpness)
    model[np.where(x > t_max)] = amplitude * np.exp( -(abs(x[np.where(x > t_max)] - t_max) / decay) ** sharpness)

    return model

def fred_flare_wrapper(params, x):
    return fred_flare(x, params)

def all_flares_fred(x, params):
    x = np.array(x)

    flare_params = [params[i:i+4] for i in range(0, len(params), 4)]
    
    sum_all_flares = [0.0] * len(x)

    for flare in flare_params:
        fit_flare = fred_flare(x, flare)
        sum_all_flares = [prev + current for prev, current in zip(sum_all_flares, fit_flare)]

    return sum_all_flares



def old_fred_flare(x, params):
    x = np.array(x)
    tau = params[0]
    rise = params[1]
    decay = params[2]
    amplitude = params[3]

    cond = x < tau  

    model = amplitude * np.sqrt(np.exp(2*(rise/decay))) * np.exp(-(rise/(x-tau))-((x-tau)/decay))
    model[np.where(cond)] = 0

    return model

def old_fred_flare_wrapper(params, x):
    return old_fred_flare(x, params)

def old_all_flares_fred(x, params):
    x = np.array(x)

    flare_params = [params[i:i+4] for i in range(0, len(params), 4)]
    
    sum_all_flares = [0.0] * len(x)

    for flare in flare_params:
        fit_flare = old_fred_flare(x, flare)
        sum_all_flares = [prev + current for prev, current in zip(sum_all_flares, fit_flare)]

    return sum_all_flares

#################################################################################
### GAUSSIAN MODEL
#################################################################################

def gaussian_flare(x, params):
    x = np.array(x)
    centre = params[0]
    height = params[1]
    width = np.abs(params[2])

    model = height * np.exp(-((x-centre)**2)/(2*(width**2)))

    return model

def gaussian_flare_wrapper(params, x):
    return gaussian_flare(x, params)

def all_flare_gauss(x, params):
    x = np.array(x)

    flare_params = [params[i:i+3] for i in range(0, len(params), 3)]
    
    sum_all_flares = [0.0] * len(x)

    for flare in flare_params:
        fit_flare = gaussian_flare(x, flare)
        sum_all_flares = [prev + current for prev, current in zip(sum_all_flares, fit_flare)]

    return sum_all_flares

#################################################################################
### SCIPY.ODR FITTING
#################################################################################

from scipy.odr import ODR, Model, RealData

def flare_fitter(data, continuum, flares, model='fred', skip_mcmc=False):
    """ 
    Flare fitting function. Takes already found flare indices and models them.

    Also runs:
      - 
    
    """
    logger.info("Fitting flares...")

    logger.debug("Calculating residuals")   

    fitted_model = broken_powerlaw(continuum['parameters'], data.time)
    residuals = data.copy()
    residuals['flux'] = data.flux - fitted_model

    flareFits = []
    flareErrs = []

    for start, peak, end in flares:

        data_flare = data.copy()
        data_flare['flux'] = np.float64(0)
        data_flare.loc[start:end, 'flux'] = residuals.loc[start:end, 'flux']

        if model == 'fred':
            flare_model = fred_flare
            model_wrapper = fred_flare_wrapper

            # Parameter estimates.
            t_max = residuals['time'].iloc[peak]
            rise = t_max - residuals['time'].iloc[start]
            decay = residuals['time'].iloc[end] - t_max
            sharpness = decay/rise
            amplitude = residuals['flux'].iloc[peak]
            input_par = [t_max, rise, decay, sharpness, amplitude]

        elif model == 'gauss':
            flare_model = gaussian_flare
            model_wrapper = gaussian_flare_wrapper

            # Parameter estimate.
            centre = residuals['time'].iloc[peak]
            height = residuals['flux'].iloc[peak]
            width = residuals['time'].iloc[peak] - residuals['time'].iloc[start]
            input_par = [centre, height, width]

        # Perform intial ODR fit.
        logger.debug(f"For flare indices {start}/{peak}/{end}:")
        odr_par, odr_err = odr_fitter(data_flare, input_par, model_wrapper)
        # odr_par = [abs(x) for x in odr_par]
        odr_stats = calculate_fit_statistics(data, flare_model, odr_par, temp_flare_shell=True)
        odr_rchisq = odr_stats['rchisq']
        logger.debug(f"ODR Par: {odr_par}")
        logger.debug(f"ODR Err: {odr_err}")

        if skip_mcmc == True:
            # Attempt mcmc fitting routine.
            try:
                mcmc_par, mcmc_err = fit_flare_mcmc(data_flare, input_par, odr_err)
                mcmc_stats = calculate_fit_statistics(data, flare_model, mcmc_par)
                mcmc_rchisq = mcmc_stats['rchisq']
                logger.debug(f"MCMC Par: {mcmc_par}")
                logger.debug(f"MCMC Err: {mcmc_err}")

                # Check for bad MCMC.
                if mcmc_rchisq == 0 or mcmc_rchisq < 0.1 or mcmc_rchisq == np.inf or mcmc_rchisq == -np.inf:
                    logger.debug(f'MCMC appears to be bad, using ODR fit for flare {start}-{end}.')
                    final_par, final_err, final_fit_statistics = odr_par, odr_err, odr_stats
                # Compare the two models.
                elif abs(odr_rchisq-1) < abs(mcmc_rchisq-1):
                    if abs(odr_rchisq) < 1.3 * abs(mcmc_rchisq-1):
                        logger.debug(f"ODR better than MCMC for flare {start}-{end}, using ODR.")
                        final_par, final_err, final_fit_statistics = odr_par, odr_err, odr_stats
                    else:
                        logger.debug(f"ODR better than MCMC fit for flare {start}-{end}, but not significantly enough.")
                # Otherwise: mcmc.
                else:
                    logger.debug("Using MCMC fit.")
                    final_par, final_err, final_fit_statistics = mcmc_par, mcmc_err, mcmc_stats
            except IndexError:
                logger.debug(f"Using ODR fit - likely gaussian curve. Not implemented in MCMC yet.")
                final_par, final_err = odr_par, odr_err
            except ValueError:
                logger.debug(f'MCMC failed - using ODR fit.')
                final_par, final_err = odr_par, odr_err
        else:
            final_par, final_err = odr_par, odr_err


        # Remove from residuals.
        fitted_flare = flare_model(data.time, final_par)
        residuals['flux'] -= fitted_flare
        
        # import matplotlib.pyplot as plt
        # import pandas as pd

        # print(residuals[['time', 'flux']])
        # plt.scatter(residuals.time, residuals.flux)
        # plt.plot(residuals.time, fitted_flare)
        # plt.semilogx()
        # plt.show()

        logger.debug("Flare complete")

        flareFits.append(list(final_par))
        flareErrs.append(list(final_err))

    logger.info("Flare fitting complete for all flares.")
    return flareFits, flareErrs


def improved_end_time(data, flare_indices, flare_par, continuum_par):
    """
    Find the end of the flares based on flare converging to continuum.
    
    Calculate the value of the fitted continuum model to the flare+continuum
    model for a series of values until the models converge. Since the FRED
    model is asymptotic to 0, we apply 'factor' to the flare+continuum, an
    small effective downshift.
    
    Parameters:
    data (pd.DataFrame): GRB lightcurve data.
    flare_indices (List[int]): a nested list of indices corresponding to flare
                               start, stop and end times.
    flare_par (List[float]): a list of the flare model parameters.
    continuum_par (List[floart]): a list of the continuum model parameters.

    Returns:
    end_index (int): the newly found index for the end of the flare.
    end_time (float): the newly found time for the end of the flare.
    """

    logger.debug(f"Running improved_end_time for flare: {flare_indices}.")
    peak, end = flare_indices[1:]
    factor = 0.99 # seems to be the most consistent value.

    if end == len(data.time) - 1:
        logger.debug("End of flare already at end of data. Returning original inputs.")
        return end, data['time'].iloc[end]

    # Calculate value for each model at current flare end.
    current_continuum = broken_powerlaw([data['time'].iloc[end]], continuum_par)[0]
    current_flare = fred_flare([data['time'].iloc[end]], flare_par)[0]
    current_flare += (current_continuum * factor)

    # Loop until the flare+continuum model drops below the continuum.
    extend = 0
    while current_flare > current_continuum and end + extend < len(data.time) -1:
        extend += 1
        logger.debug(f"Conditions not yet met -- extending to {extend}")
        current_continuum = broken_powerlaw([data['time'].iloc[end+extend]], continuum_par)[0]
        current_flare = fred_flare([data['time'].iloc[end+extend]], flare_par)[0]
        current_flare += (current_continuum * factor)
    end_index = end + extend
    logger.debug(f"Conditions met! Calculating intercept.")
        
    # Calculate the intercept time.
    search_time = np.logspace(
        np.log10(data['time'].iloc[peak+1]), 
        np.log10(data['time'].iloc[end+extend+1]),
        num = 1000)
    
    continuum_model = broken_powerlaw(search_time, continuum_par)
    flare_model = fred_flare(search_time, flare_par) + (continuum_model * factor)
    end_time, _ = intersection(search_time, continuum_model, search_time, flare_model)

    # If no intersect found:
    if len(end_time) == 0:
        # print("NO TIME FOUND")
        end_time = data['time'].iloc[end]

    elif len(end_time) > 1:
        logger.warning("Multiple intercepts found for this burst - the flare \
                       may be incorrect or modelled badly.")
        end_time = end_time[-1]
        # print("WEVE ENTERED GREATER THAN ONE")
        # import matplotlib.pyplot as plt
        # plt.plot(search_time, continuum_model, color='tab:orange')
        # plt.plot(search_time, flare_model, color='r')
        # plt.scatter(data.time, data.flux, color='b', marker='.')
        # plt.loglog()
        # plt.show()
        # raise ValueError("It appears two intercepts were found!")
        ## This can occur when the found flare isn't really a flare. Noisy data
        ## etc will cause the flare to be modelled wrong, and so multiple
        ## intersections happen. I can't think of a scenario otherwise? There
        ## should only ever be 2, so just select the later one in this case.
        ## Maybe I can remove this check with better flare finding?
    else:
        end_time = end_time[0]
        # print("WEVE NOT NOT NOT")
    logger.debug(f"Found new end information: index {end_index}, end_time {end_time}.")
    # print(end_time)
    return end_index, end_time


#################################################################################
### FITTING METHODS
#################################################################################

def odr_fitter(data, inputpar, model_wrapper):
    data = RealData(data.time, data.flux, data.time_perr, data.flux_perr)  
    model = Model(model_wrapper)
    odr = ODR(data, model, beta0=inputpar)

    odr.set_job(fit_type=0)
    output = odr.run()

    if output.info != 1:
        i = 1
        while output.info != 1 and i < 100:
            output = odr.restart()
            i += 1

    return output.beta, output.sd_beta

def fit_flare_mcmc(data, init_param, init_err):

    ndim = len(init_param)
    nwalkers = 30
    nsteps = 300

    p0 = np.zeros((nwalkers, ndim))

    guess_tstart = init_param[0]
    std_tstart = init_param[0]
    p0[:, 0] = guess_tstart + std_tstart * np.random.randn(nwalkers)

    guess_rise = init_param[1]
    std_rise = init_param[1]
    p0[:, 1] = guess_rise + std_rise * np.random.randn(nwalkers)

    guess_decay = init_param[2]
    std_decay = init_param[2]
    p0[:, 2] = guess_decay + std_decay * np.random.randn(nwalkers)

    guess_amplitude = init_param[3]
    std_ampltiude = init_param[3]
    p0[:, 3] = guess_amplitude + std_ampltiude * np.random.randn(nwalkers)

    logger.debug("Running flare MCMC...")

    sampler = emcee.EnsembleSampler(nwalkers, ndim, fl_log_posterior, \
        args=(data.time, data.flux, data.time_perr, data.flux_perr))
    sampler.run_mcmc(p0, nsteps)

    burnin = 50

    samples = sampler.chain[:, burnin:, :].reshape(-1, ndim)

    fitted_par = list(map(lambda v: np.median(v), samples.T))
    fitted_err = list(map(lambda v: np.std(v), samples.T))

    logger.debug("MCMC run completed.")

    return fitted_par, fitted_err

def fl_log_likelihood(params, x, y, x_err, y_err):
    model = fred_flare(x, params)
    chisq = np.sum(( (y-model)**2) / ((y_err)**2)) 
    log_likelihood = -0.5 * np.sum(chisq + np.log(2 * np.pi * y_err**2))
    return log_likelihood   

def fl_log_prior(params, TIME_END):

    tau = params[0]
    rise = params[1]
    decay = params[2]
    amplitude = params[3]

    if not (tau > 0) and (tau < TIME_END):
        return -np.inf

    if rise < 0:
        return -np.inf
    if rise > TIME_END:
        return -np.inf

    if decay < 0:
        return -np.inf
    if decay > TIME_END:
        return -np.inf

    if amplitude < 0:
        return -np.inf

    return 0.0

def fl_log_posterior(params, x, y, x_err, y_err):
    lp = fl_log_prior(params, x.iloc[-1])
    ll = fl_log_likelihood(params, x, y, x_err, y_err)
    if not np.isfinite(lp):
        return -np.inf
    return lp + ll